"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from amds.types import BaseModel
from amds.utils import FieldMetadata, QueryParamMetadata
from typing import Dict, List, Optional
from typing_extensions import Annotated, NotRequired, TypedDict


class GetTunnelRequestTypedDict(TypedDict):
    tunnel_id: NotRequired[str]


class GetTunnelRequest(BaseModel):
    tunnel_id: Annotated[
        Optional[str],
        FieldMetadata(query=QueryParamMetadata(style="form", explode=True)),
    ] = None


class GetTunnelDataTypedDict(TypedDict):
    token: NotRequired[str]
    tunnel_id: NotRequired[str]


class GetTunnelData(BaseModel):
    token: Optional[str] = None

    tunnel_id: Optional[str] = None


class GetTunnelResponseBodyTypedDict(TypedDict):
    r"""Get Tunnel Token"""

    data: NotRequired[GetTunnelDataTypedDict]
    status: NotRequired[str]


class GetTunnelResponseBody(BaseModel):
    r"""Get Tunnel Token"""

    data: Optional[GetTunnelData] = None

    status: Optional[str] = None


class GetTunnelResponseTypedDict(TypedDict):
    headers: Dict[str, List[str]]
    result: GetTunnelResponseBodyTypedDict


class GetTunnelResponse(BaseModel):
    headers: Dict[str, List[str]]

    result: GetTunnelResponseBody



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from .addintegratedserverop import (
    AddIntegratedServerData,
    AddIntegratedServerDataTypedDict,
    AddIntegratedServerRequestBody,
    AddIntegratedServerRequestBodyTypedDict,
    AddIntegratedServerResponse,
    AddIntegratedServerResponseBody,
    AddIntegratedServerResponseBodyTypedDict,
    AddIntegratedServerResponseTypedDict,
)
from .apierror import APIError
from .claude35haikuop import (
    Claude35HaikuMessages,
    Claude35HaikuMessagesTypedDict,
    Claude35HaikuRequest,
    Claude35HaikuRequestBody,
    Claude35HaikuRequestBodyTypedDict,
    Claude35HaikuRequestTypedDict,
    Claude35HaikuResponse,
    Claude35HaikuResponseTypedDict,
)
from .claude35sonnetop import (
    Claude35SonnetMessages,
    Claude35SonnetMessagesTypedDict,
    Claude35SonnetRequest,
    Claude35SonnetRequestBody,
    Claude35SonnetRequestBodyTypedDict,
    Claude35SonnetRequestTypedDict,
    Claude35SonnetResponse,
    Claude35SonnetResponseTypedDict,
)
from .createserverop import (
    CreateServerRequestBody,
    CreateServerRequestBodyTypedDict,
    CreateServerResponse,
    CreateServerResponseBody,
    CreateServerResponseBodyTypedDict,
    CreateServerResponseTypedDict,
)
from .createtunnelop import (
    CreateTunnelData,
    CreateTunnelDataTypedDict,
    CreateTunnelRequestBody,
    CreateTunnelRequestBodyTypedDict,
    CreateTunnelResponse,
    CreateTunnelResponseBody,
    CreateTunnelResponseBodyTypedDict,
    CreateTunnelResponseTypedDict,
)
from .deleteintegratedserverop import (
    DeleteIntegratedServerRequest,
    DeleteIntegratedServerRequestTypedDict,
    DeleteIntegratedServerResponse,
    DeleteIntegratedServerResponseBody,
    DeleteIntegratedServerResponseBodyTypedDict,
    DeleteIntegratedServerResponseTypedDict,
)
from .deleteserverop import (
    DeleteServerRequest,
    DeleteServerRequestTypedDict,
    DeleteServerResponse,
    DeleteServerResponseBody,
    DeleteServerResponseBodyTypedDict,
    DeleteServerResponseTypedDict,
)
from .deletetunnelop import (
    DeleteTunnelRequest,
    DeleteTunnelRequestTypedDict,
    DeleteTunnelResponse,
    DeleteTunnelResponseBody,
    DeleteTunnelResponseBodyTypedDict,
    DeleteTunnelResponseTypedDict,
)
from .getcomputeop import (
    Availability,
    AvailabilityTypedDict,
    GetComputeCompute,
    GetComputeComputeTypedDict,
    GetComputeResponse,
    GetComputeResponseBody,
    GetComputeResponseBodyTypedDict,
    GetComputeResponseTypedDict,
)
from .getenvironmentsop import (
    GetEnvironmentsEnvironments,
    GetEnvironmentsEnvironmentsTypedDict,
    GetEnvironmentsResponse,
    GetEnvironmentsResponseBody,
    GetEnvironmentsResponseBodyTypedDict,
    GetEnvironmentsResponseTypedDict,
)
from .getfileop import (
    Content,
    ContentTypedDict,
    GetFileRequest,
    GetFileRequestTypedDict,
    GetFileResponse,
    GetFileResponseBody,
    GetFileResponseBodyTypedDict,
    GetFileResponseTypedDict,
)
from .getintegratedserversop import (
    Data,
    DataTypedDict,
    GetIntegratedServersResponse,
    GetIntegratedServersResponseBody,
    GetIntegratedServersResponseBodyTypedDict,
    GetIntegratedServersResponseTypedDict,
)
from .getserversop import (
    GetServersResponse,
    GetServersResponseBody,
    GetServersResponseBodyTypedDict,
    GetServersResponseTypedDict,
    GetServersServers,
    GetServersServersTypedDict,
)
from .gettunnelop import (
    GetTunnelData,
    GetTunnelDataTypedDict,
    GetTunnelRequest,
    GetTunnelRequestTypedDict,
    GetTunnelResponse,
    GetTunnelResponseBody,
    GetTunnelResponseBodyTypedDict,
    GetTunnelResponseTypedDict,
)
from .gpt4ominiop import (
    Gpt4oMiniRequest,
    Gpt4oMiniRequestBody,
    Gpt4oMiniRequestBodyTypedDict,
    Gpt4oMiniRequestTypedDict,
    Gpt4oMiniResponse,
    Gpt4oMiniResponseTypedDict,
    Messages,
    MessagesTypedDict,
)
from .gpt4oop import (
    Gpt4oMessages,
    Gpt4oMessagesTypedDict,
    Gpt4oRequest,
    Gpt4oRequestBody,
    Gpt4oRequestBodyTypedDict,
    Gpt4oRequestTypedDict,
    Gpt4oResponse,
    Gpt4oResponseTypedDict,
)
from .gpt4op import (
    Gpt4Messages,
    Gpt4MessagesTypedDict,
    Gpt4Request,
    Gpt4RequestBody,
    Gpt4RequestBodyTypedDict,
    Gpt4RequestTypedDict,
    Gpt4Response,
    Gpt4ResponseTypedDict,
)
from .runcodeop import (
    Metadata,
    MetadataTypedDict,
    Outputs,
    OutputsTypedDict,
    RunCodeContent,
    RunCodeContentTypedDict,
    RunCodeRequest,
    RunCodeRequestBody,
    RunCodeRequestBodyTypedDict,
    RunCodeRequestTypedDict,
    RunCodeResponse,
    RunCodeResponseBody,
    RunCodeResponseBodyTypedDict,
    RunCodeResponseTypedDict,
)
from .security import Security, SecurityTypedDict
from .startserverop import (
    StartServerRequest,
    StartServerRequestBody,
    StartServerRequestBodyTypedDict,
    StartServerRequestTypedDict,
    StartServerResponse,
    StartServerResponseBody,
    StartServerResponseBodyTypedDict,
    StartServerResponseTypedDict,
)
from .stopserverop import (
    StopServerRequest,
    StopServerRequestTypedDict,
    StopServerResponse,
    StopServerResponseBody,
    StopServerResponseBodyTypedDict,
    StopServerResponseTypedDict,
)
from .uploadfileop import (
    UploadFileContent,
    UploadFileContentTypedDict,
    UploadFileRequest,
    UploadFileRequestBody,
    UploadFileRequestBodyTypedDict,
    UploadFileRequestTypedDict,
    UploadFileResponse,
    UploadFileResponseBody,
    UploadFileResponseBodyTypedDict,
    UploadFileResponseTypedDict,
)


__all__ = [
    "APIError",
    "AddIntegratedServerData",
    "AddIntegratedServerDataTypedDict",
    "AddIntegratedServerRequestBody",
    "AddIntegratedServerRequestBodyTypedDict",
    "AddIntegratedServerResponse",
    "AddIntegratedServerResponseBody",
    "AddIntegratedServerResponseBodyTypedDict",
    "AddIntegratedServerResponseTypedDict",
    "Availability",
    "AvailabilityTypedDict",
    "Claude35HaikuMessages",
    "Claude35HaikuMessagesTypedDict",
    "Claude35HaikuRequest",
    "Claude35HaikuRequestBody",
    "Claude35HaikuRequestBodyTypedDict",
    "Claude35HaikuRequestTypedDict",
    "Claude35HaikuResponse",
    "Claude35HaikuResponseTypedDict",
    "Claude35SonnetMessages",
    "Claude35SonnetMessagesTypedDict",
    "Claude35SonnetRequest",
    "Claude35SonnetRequestBody",
    "Claude35SonnetRequestBodyTypedDict",
    "Claude35SonnetRequestTypedDict",
    "Claude35SonnetResponse",
    "Claude35SonnetResponseTypedDict",
    "Content",
    "ContentTypedDict",
    "CreateServerRequestBody",
    "CreateServerRequestBodyTypedDict",
    "CreateServerResponse",
    "CreateServerResponseBody",
    "CreateServerResponseBodyTypedDict",
    "CreateServerResponseTypedDict",
    "CreateTunnelData",
    "CreateTunnelDataTypedDict",
    "CreateTunnelRequestBody",
    "CreateTunnelRequestBodyTypedDict",
    "CreateTunnelResponse",
    "CreateTunnelResponseBody",
    "CreateTunnelResponseBodyTypedDict",
    "CreateTunnelResponseTypedDict",
    "Data",
    "DataTypedDict",
    "DeleteIntegratedServerRequest",
    "DeleteIntegratedServerRequestTypedDict",
    "DeleteIntegratedServerResponse",
    "DeleteIntegratedServerResponseBody",
    "DeleteIntegratedServerResponseBodyTypedDict",
    "DeleteIntegratedServerResponseTypedDict",
    "DeleteServerRequest",
    "DeleteServerRequestTypedDict",
    "DeleteServerResponse",
    "DeleteServerResponseBody",
    "DeleteServerResponseBodyTypedDict",
    "DeleteServerResponseTypedDict",
    "DeleteTunnelRequest",
    "DeleteTunnelRequestTypedDict",
    "DeleteTunnelResponse",
    "DeleteTunnelResponseBody",
    "DeleteTunnelResponseBodyTypedDict",
    "DeleteTunnelResponseTypedDict",
    "GetComputeCompute",
    "GetComputeComputeTypedDict",
    "GetComputeResponse",
    "GetComputeResponseBody",
    "GetComputeResponseBodyTypedDict",
    "GetComputeResponseTypedDict",
    "GetEnvironmentsEnvironments",
    "GetEnvironmentsEnvironmentsTypedDict",
    "GetEnvironmentsResponse",
    "GetEnvironmentsResponseBody",
    "GetEnvironmentsResponseBodyTypedDict",
    "GetEnvironmentsResponseTypedDict",
    "GetFileRequest",
    "GetFileRequestTypedDict",
    "GetFileResponse",
    "GetFileResponseBody",
    "GetFileResponseBodyTypedDict",
    "GetFileResponseTypedDict",
    "GetIntegratedServersResponse",
    "GetIntegratedServersResponseBody",
    "GetIntegratedServersResponseBodyTypedDict",
    "GetIntegratedServersResponseTypedDict",
    "GetServersResponse",
    "GetServersResponseBody",
    "GetServersResponseBodyTypedDict",
    "GetServersResponseTypedDict",
    "GetServersServers",
    "GetServersServersTypedDict",
    "GetTunnelData",
    "GetTunnelDataTypedDict",
    "GetTunnelRequest",
    "GetTunnelRequestTypedDict",
    "GetTunnelResponse",
    "GetTunnelResponseBody",
    "GetTunnelResponseBodyTypedDict",
    "GetTunnelResponseTypedDict",
    "Gpt4Messages",
    "Gpt4MessagesTypedDict",
    "Gpt4Request",
    "Gpt4RequestBody",
    "Gpt4RequestBodyTypedDict",
    "Gpt4RequestTypedDict",
    "Gpt4Response",
    "Gpt4ResponseTypedDict",
    "Gpt4oMessages",
    "Gpt4oMessagesTypedDict",
    "Gpt4oMiniRequest",
    "Gpt4oMiniRequestBody",
    "Gpt4oMiniRequestBodyTypedDict",
    "Gpt4oMiniRequestTypedDict",
    "Gpt4oMiniResponse",
    "Gpt4oMiniResponseTypedDict",
    "Gpt4oRequest",
    "Gpt4oRequestBody",
    "Gpt4oRequestBodyTypedDict",
    "Gpt4oRequestTypedDict",
    "Gpt4oResponse",
    "Gpt4oResponseTypedDict",
    "Messages",
    "MessagesTypedDict",
    "Metadata",
    "MetadataTypedDict",
    "Outputs",
    "OutputsTypedDict",
    "RunCodeContent",
    "RunCodeContentTypedDict",
    "RunCodeRequest",
    "RunCodeRequestBody",
    "RunCodeRequestBodyTypedDict",
    "RunCodeRequestTypedDict",
    "RunCodeResponse",
    "RunCodeResponseBody",
    "RunCodeResponseBodyTypedDict",
    "RunCodeResponseTypedDict",
    "Security",
    "SecurityTypedDict",
    "StartServerRequest",
    "StartServerRequestBody",
    "StartServerRequestBodyTypedDict",
    "StartServerRequestTypedDict",
    "StartServerResponse",
    "StartServerResponseBody",
    "StartServerResponseBodyTypedDict",
    "StartServerResponseTypedDict",
    "StopServerRequest",
    "StopServerRequestTypedDict",
    "StopServerResponse",
    "StopServerResponseBody",
    "StopServerResponseBodyTypedDict",
    "StopServerResponseTypedDict",
    "UploadFileContent",
    "UploadFileContentTypedDict",
    "UploadFileRequest",
    "UploadFileRequestBody",
    "UploadFileRequestBodyTypedDict",
    "UploadFileRequestTypedDict",
    "UploadFileResponse",
    "UploadFileResponseBody",
    "UploadFileResponseBodyTypedDict",
    "UploadFileResponseTypedDict",
]



import click
import os
import sys
import json
import subprocess
import shutil
import tempfile
import webbrowser
import requests
import time
from pathlib import Path
import socket
import random
import string
from urllib.parse import urlparse
import threading
import signal
import hashlib
import re
import queue
from ..utils import print_json


def create_jupyter_config(
    config_dir,
    allow_origin="https://*.amdatascience.com",
    disable_sudo=False,
):
    """
    Create a Jupyter config file with CORS settings

    Args:
        config_dir: Directory to create the config in
        allow_origin: CORS allow-origin setting
        disable_sudo: Whether to disable sudo/root access

    Returns:
        Path to the created config file
    """
    # Create the config directory if it doesn't exist
    os.makedirs(config_dir, exist_ok=True)

    # Create a jupyter_server_config.py file with CORS configuration
    config_file = os.path.join(config_dir, "jupyter_server_config.py")

    sudo_config = ""
    if disable_sudo:
        sudo_config = """
# Disable sudo/root access
c.ServerApp.allow_root = False
c.ServerApp.allow_sudo = False
"""

    with open(config_file, "w") as f:
        f.write(
            f"""
# Configuration file for jupyter-server.

c = get_config()

# Configure CORS settings
c.ServerApp.allow_origin = '{allow_origin}'
c.ServerApp.allow_credentials = True
c.ServerApp.allow_methods = ['*']
c.ServerApp.allow_headers = ['Content-Type', 'Authorization', 'X-Requested-With', 
                            'X-XSRFToken', 'ngrok-skip-browser-warning', 'Origin', 
                            'Accept', 'Cache-Control', 'X-Requested-With', '*']
{sudo_config}
"""
        )

    return config_file


def is_port_in_use(port):
    """Check if a port is already in use"""
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        return s.connect_ex(("localhost", port)) == 0


def find_free_port(start_port, max_attempts=10):
    """Find a free port starting from start_port"""
    port = start_port
    attempts = 0

    while attempts < max_attempts:
        if not is_port_in_use(port):
            return port
        port += 1
        attempts += 1

    return None


def check_jupyter_health(jupyter_url, token):
    """Check if Jupyter server is responding properly"""
    try:
        # Extract base URL without token
        base_url = jupyter_url.split("?")[0]
        # Construct the API URL to check server status
        api_url = f"{base_url}/api/status"
        headers = {}
        if token:
            headers["Authorization"] = f"token {token}"

        response = requests.get(api_url, headers=headers, timeout=5)
        return response.status_code == 200
    except Exception:
        return False


def check_tunnel_health(tunnel_url, provider):
    """Check if tunnel is responding properly"""
    try:
        headers = {}
        if provider == "ngrok":
            # Add parameter to bypass ngrok browser warning
            check_url = f"{tunnel_url}?ngrok-skip-browser-warning=true"
            headers["ngrok-skip-browser-warning"] = "true"
        else:
            check_url = tunnel_url

        response = requests.get(check_url, timeout=5, headers=headers)
        return response.status_code < 400
    except Exception:
        return False


def start_jupyter_server(directory, port, jupyter_config):
    """
    Start Jupyter Lab server and extract token

    Returns:
        tuple: (jupyter_process, jupyter_url, token)
    """
    click.echo(f"Starting Jupyter Lab on port {port}...")
    jupyter_cmd = [
        "jupyter",
        "lab",
        f"--port={port}",
        "--no-browser",
        f"--notebook-dir={directory}",
        "--ip=0.0.0.0",
        f"--config={jupyter_config}",
    ]

    # Start Jupyter process
    jupyter_process = subprocess.Popen(
        jupyter_cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        text=True,
        bufsize=1,
    )

    # Wait for Jupyter to start and get the token
    token = None
    jupyter_url = None
    for line in iter(jupyter_process.stderr.readline, ""):
        if "http://localhost:" in line or "http://127.0.0.1:" in line:
            # Extract just the URL part, removing the log prefix
            url_part = re.search(r"(http://[^\s]+)", line)
            if url_part:
                jupyter_url = url_part.group(0).strip()
                token = jupyter_url.split("token=")[-1].strip()
                break
            else:
                # Fallback to old method if regex fails
                jupyter_url = line.strip().split("or ")[-1].strip()
                token = jupyter_url.split("token=")[-1].strip()
                break

        sys.stderr.write(line)

        if not token and "ERROR" in line:
            click.secho(
                "Error starting Jupyter Lab. Please check the logs above.",
                fg="red",
                bold=True,
            )
            jupyter_process.terminate()
            sys.exit(1)

    if not token:
        click.secho("Error: Could not get Jupyter token. Exiting.", fg="red", bold=True)
        jupyter_process.terminate()
        sys.exit(1)

    click.secho("✓ Jupyter Lab started successfully", fg="green", bold=True)

    # Short delay to ensure all output is captured
    time.sleep(0.5)

    return jupyter_process, jupyter_url, token


def setup_tunnel(client, port, app_port=None, server_name=None):
    """
    Set up Cloudflare tunnel via API

    Returns:
        dict: Tunnel information or None if failed
    """
    click.secho(
        "Setting up tunnel via API...",
        fg="blue",
        bold=True,
    )
    try:
        with client as c:
            # Call the tunnels API to get a tunnel token
            request_data = {
                "jupyter_port": port,
            }

            # Add app port if specified
            if app_port:
                request_data["app_port"] = app_port

            # Add server name if we want a specific one
            if not server_name:
                server_name = f"local-{int(time.time())}"
            request_data["server_name"] = server_name

            # Call the API to create the tunnel
            try:
                result = c.tunnels.create(request=request_data)

                # Extract tunnel data from the response
                if hasattr(result, "result") and hasattr(result.result, "data"):
                    tunnel_data = result.result.data
                    cloudflare_token = tunnel_data.token
                    tunnel_url = tunnel_data.jupyter_url
                    tunnel_name = tunnel_data.name
                    tunnel_id = getattr(tunnel_data, "id", None)
                    app_tunnel_url = getattr(tunnel_data, "app_url", None)

                    # Install the tunnel service
                    click.secho("Installing tunnel service...", fg="blue")
                    install_cmd = [
                        "sudo",
                        "cloudflared",
                        "service",
                        "install",
                        cloudflare_token,
                    ]

                    # Run the install command
                    subprocess.run(install_cmd, check=True)
                    click.secho(
                        "✓ tunnel service installed",
                        fg="green",
                        bold=True,
                    )

                    return {
                        "tunnel_url": tunnel_url,
                        "app_tunnel_url": app_tunnel_url,
                        "cloudflare_token": cloudflare_token,
                        "tunnel_name": tunnel_name,
                        "tunnel_id": tunnel_id,
                    }
                else:
                    click.secho(
                        "Warning: Invalid response from tunnels API.",
                        fg="yellow",
                    )
                    return None

            except Exception as e:
                click.secho(
                    f"Error creating tunnel: {str(e)}",
                    fg="yellow",
                )
                return None
    except Exception as e:
        click.secho(f"Error setting up tunnel: {str(e)}", fg="yellow")
        return None


def setup_temporary_tunnel(port, app_port=None):
    """
    Set up temporary Cloudflare tunnel

    Returns:
        dict: Tunnel information
    """
    # Start Cloudflare tunnel
    click.secho(
        f"Starting temporary tunnel to port {port}...",
        fg="blue",
        bold=True,
    )
    cloudflare_cmd = [
        "cloudflared",
        "tunnel",
        "--url",
        f"http://localhost:{port}",
    ]

    tunnel_process = subprocess.Popen(
        cloudflare_cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        text=True,
        bufsize=1,
    )

    # Wait for cloudflare tunnel to start and get the public URL
    tunnel_url = None
    app_tunnel_url = None
    app_tunnel_process = None

    # Define reader function for output stream
    def reader(stream, queue):
        for line in iter(stream.readline, ""):
            queue.put(line)
        stream.close()

    # Create a queue for the output lines
    output_queue = queue.Queue()

    # Start reader threads for both stdout and stderr
    threading.Thread(
        target=reader, args=(tunnel_process.stdout, output_queue), daemon=True
    ).start()
    threading.Thread(
        target=reader, args=(tunnel_process.stderr, output_queue), daemon=True
    ).start()

    # Process the output lines to find tunnel URL
    tunnel_url = find_tunnel_url(output_queue, tunnel_process)

    if not tunnel_url:
        click.secho(
            "Error: Could not get tunnel URL. Exiting.",
            fg="red",
            bold=True,
        )
        click.echo(
            "Try running 'cloudflared tunnel --url http://localhost:8888' manually to check for errors."
        )
        tunnel_process.terminate()
        return None

    click.secho(f"Tunnel created at {tunnel_url}", fg="green")

    # Start additional tunnel for app port if specified
    if app_port:
        app_tunnel_url, app_tunnel_process = setup_app_tunnel(app_port)

    return {
        "tunnel_url": tunnel_url,
        "tunnel_process": tunnel_process,
        "app_tunnel_url": app_tunnel_url,
        "app_tunnel_process": app_tunnel_process,
    }


def find_tunnel_url(output_queue, process, timeout=60):
    """Extract tunnel URL from process output queue"""
    tunnel_url = None
    start_time = time.time()

    while time.time() < start_time + timeout:
        try:
            # Get line with timeout to avoid blocking forever
            line = output_queue.get(timeout=1)
            sys.stdout.write(line)

            # Look for the cloudflare tunnel URL
            if "https://" in line and "trycloudflare.com" in line:
                match = re.search(r"https://[^\s|]+", line)
                if match:
                    tunnel_url = match.group(0).strip()
                    break

            # Also check for URLs without table formatting
            if "https://" in line:
                match = re.search(r"https://[^ |\n\r]+\.trycloudflare\.com", line)
                if match:
                    tunnel_url = match.group(0).strip()
                    break

            if "error" in line.lower():
                click.echo("Error starting tunnel. Please check the logs above.")
                process.terminate()
                return None

        except queue.Empty:
            # No output for a second, check if process is still alive
            if process.poll() is not None:
                # Process has ended
                break
            continue

    return tunnel_url


def setup_app_tunnel(app_port):
    """Set up additional tunnel for application port"""
    click.secho(
        f"Starting additional tunnel for application port {app_port}...",
        fg="blue",
        bold=True,
    )
    app_cloudflare_cmd = [
        "cloudflared",
        "tunnel",
        "--url",
        f"http://localhost:{app_port}",
    ]

    app_tunnel_process = subprocess.Popen(
        app_cloudflare_cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        text=True,
        bufsize=1,
    )

    # Create a queue for the output lines
    app_output_queue = queue.Queue()

    # Define reader function
    def reader(stream, queue):
        for line in iter(stream.readline, ""):
            queue.put(line)
        stream.close()

    # Start reader threads for app tunnel
    threading.Thread(
        target=reader,
        args=(app_tunnel_process.stdout, app_output_queue),
        daemon=True,
    ).start()
    threading.Thread(
        target=reader,
        args=(app_tunnel_process.stderr, app_output_queue),
        daemon=True,
    ).start()

    app_tunnel_url = find_tunnel_url(app_output_queue, app_tunnel_process)

    if app_tunnel_url:
        click.secho(
            f"Additional tunnel for app port created at {app_tunnel_url}",
            fg="green",
        )
    else:
        click.secho(
            "Warning: Could not create additional tunnel for app port. Continuing without app tunnel.",
            fg="yellow",
        )

    return app_tunnel_url, app_tunnel_process


def register_with_dashboard(
    client,
    server_name,
    token,
    tunnel_url,
    app_tunnel_url=None,
    app_port=None,
    thumbnail=None,
):
    """
    Register server with dashboard

    Returns:
        str: Registered server name or None if failed
    """
    click.echo("Uploading information to dashboard.amdatascience.com...")

    try:
        with client as c:
            # Format data according to the expected API format
            try:
                # Add the integrated server using the proper API method
                request_data = {
                    "environment": "local",
                    "server_name": server_name,
                    "token": token,
                    "url": tunnel_url,
                }

                # Add app port forward URL if available
                if app_tunnel_url:
                    request_data["port_forward_url"] = app_tunnel_url

                # Add the app port number if available
                if app_port:
                    request_data["port"] = app_port

                # Add custom thumbnail if provided, otherwise use default
                if thumbnail:
                    request_data["thumb"] = thumbnail
                else:
                    # Use Jupyter's favicon as default thumbnail
                    request_data["thumb"] = f"/images/compute_icons/amds.svg"

                # Get the API key from the client or environment
                api_key_to_use = get_api_key_from_client(client)

                if api_key_to_use:
                    from amds import Amds

                    fresh_client = Amds(api_key=api_key_to_use)
                    # Call the add method directly with the fresh client
                    result = fresh_client.integrated_servers.add(request=request_data)
                else:
                    click.echo(
                        "Falling back to original client for server registration"
                    )
                    # Try the original client but with explicit request parameter
                    result = c.integrated_servers.add(request=request_data)

                click.secho(f"Server registered as '{server_name}'", fg="green")
                dashboard_url = "https://dashboard.amdatascience.com"
                click.echo(
                    f"View in dashboard: {click.style(dashboard_url, fg='bright_blue')}"
                )
                return server_name

            except Exception as e:
                click.echo(f"Warning: Could not register with dashboard: {str(e)}")
                try:
                    # Try alternative method as last resort - direct call
                    result = c.integrated_servers.add(**request_data)
                    return server_name
                except:
                    return None
    except Exception as e:
        click.echo(f"Warning: Failed to connect to dashboard: {str(e)}")
        return None


def get_api_key_from_client(client):
    """Extract API key from client object"""
    if hasattr(client, "api_key"):
        return client.api_key
    elif hasattr(client, "sdk_configuration") and hasattr(
        client.sdk_configuration, "security"
    ):
        return client.sdk_configuration.security.api_key
    elif hasattr(client, "_api_key"):
        return client._api_key
    else:
        return os.environ.get("AMDS_API_KEY")


def cleanup_cloudflare_service():
    """Uninstall Cloudflare tunnel service"""
    try:
        click.echo("Uninstalling tunnel service...")
        # Run the uninstall command (this may need sudo)
        subprocess.run(["sudo", "cloudflared", "service", "uninstall"], check=False)
    except Exception as e:
        click.echo(f"Warning: Failed to uninstall tunnel service: {str(e)}")
        click.echo(
            "You may need to manually uninstall it with: sudo cloudflared service uninstall"
        )


def cleanup_tunnel_api(client, tunnel_id, api_key=None):
    """Clean up Cloudflare tunnel via API"""
    click.echo(f"Cleaning up tunnel via API...")
    try:
        # Import the SDK if needed
        from amds import Amds

        # Get the API key - prioritize the directly provided one
        cleanup_api_key = api_key or get_api_key_from_client(client)

        if cleanup_api_key:
            # Create a new client without using a context manager
            cleanup_client = Amds(api_key=cleanup_api_key)
            # Call delete tunnel API
            try:
                result = cleanup_client.tunnels.delete(tunnel_id=tunnel_id)
                click.echo(f"Tunnel with ID '{tunnel_id}' deleted")
            except Exception as e1:
                click.echo(f"Tunnel deletion failed: {str(e1)}")
                click.echo("The tunnel may need to be manually removed.")
        else:
            click.echo(
                "Warning: No API key available for tunnel cleanup. The tunnel may need to be manually removed."
            )
    except Exception as e:
        click.echo(f"Warning: Failed to connect to API for tunnel cleanup: {str(e)}")


def cleanup_dashboard_integration(client, registered_server_name, api_key=None):
    """Clean up dashboard integration by deleting the server record"""
    click.echo(f"Cleaning up dashboard integration...")
    try:
        # Import the SDK if needed
        from amds import Amds

        # Get the API key - prioritize the directly provided one
        cleanup_api_key = api_key or get_api_key_from_client(client)

        if cleanup_api_key:
            # Create a new client without using a context manager
            cleanup_client = Amds(api_key=cleanup_api_key)
            # Call delete directly without using a context manager
            try:
                result = cleanup_client.integrated_servers.delete(
                    server_name=registered_server_name
                )
                click.echo(
                    f"Server '{registered_server_name}' unregistered from dashboard"
                )
            except Exception as e1:
                click.echo(f"Deletion failed: {str(e1)}")
                click.echo(
                    "The server may need to be manually removed from the dashboard."
                )
        else:
            click.echo(
                "Warning: No API key available for cleanup. The server may need to be manually removed."
            )
    except Exception as e:
        click.echo(f"Warning: Failed to connect to dashboard for cleanup: {str(e)}")


def start_health_check_thread(jupyter_url, token, tunnel_url, app_tunnel_url=None):
    """Start a thread to periodically check health of services"""

    def health_checker():
        """Periodically check if services are still running"""
        while True:
            time.sleep(30)  # Check every 30 seconds
            try:
                # Check if Jupyter is still responsive
                if not check_jupyter_health(jupyter_url, token):
                    click.secho(
                        "Warning: Jupyter server is not responding! The application may be experiencing issues.",
                        fg="yellow",
                        bold=True,
                    )

                # Check if tunnel is still connected
                if not check_tunnel_health(tunnel_url, "cloudflare"):
                    click.secho(
                        "Warning: Tunnel may be down! Public URL may not be accessible.",
                        fg="yellow",
                        bold=True,
                    )

                # Check app tunnel health if it exists
                if app_tunnel_url and not check_tunnel_health(
                    app_tunnel_url, "cloudflare"
                ):
                    click.secho(
                        "Warning: App tunnel may be down! App URL may not be accessible.",
                        fg="yellow",
                        bold=True,
                    )
            except Exception as e:
                # Log exceptions but don't crash the monitoring thread
                click.secho(
                    f"Warning: Health check encountered an error: {str(e)}",
                    fg="yellow",
                )

    # Start health check in a daemon thread
    health_thread = threading.Thread(target=health_checker, daemon=True)
    health_thread.start()
    return health_thread


def display_output(
    output_format,
    jupyter_url,
    tunnel_url,
    token,
    app_tunnel_url=None,
    registered_server_name=None,
    cloudflare_token=None,
):
    """Display output information based on user preference"""
    if output_format == "standard":
        click.echo("\n" + "=" * 60)
        click.secho("Alph Editor is connected with local server", fg="green", bold=True)
        click.echo("=" * 60)
        click.echo("Local URL:        " + click.style(jupyter_url, fg="bright_blue"))
        click.echo("Public URL:       " + click.style(tunnel_url, fg="bright_blue"))
        if app_tunnel_url:
            click.echo(
                "App URL:          " + click.style(app_tunnel_url, fg="bright_blue")
            )
        if registered_server_name:
            alph_url = f"https://dashboard.amdatascience.com/alph-editor/{registered_server_name}"
            click.echo("Alph Editor URL:  " + click.style(alph_url, fg="bright_blue"))
        click.echo("=" * 60)
        click.echo("\nPress Ctrl+C to stop the server...\n")
    elif output_format == "minimal":
        click.echo("Local URL:        " + jupyter_url)
        click.echo("Public URL:       " + tunnel_url)
        if app_tunnel_url:
            click.echo("App URL:          " + app_tunnel_url)
        if registered_server_name:
            alph_url = f"https://dashboard.amdatascience.com/alph-editor/{registered_server_name}"
            click.echo("Alph Editor URL:  " + alph_url)
        click.echo("\nPress Ctrl+C to stop the server...")
    elif output_format == "json":
        output = {
            "local_url": jupyter_url,
            "public_url": tunnel_url,
            "token": token,
        }
        if app_tunnel_url:
            output["app_url"] = app_tunnel_url
        if registered_server_name:
            output["alph_url"] = (
                f"https://dashboard.amdatascience.com/alph-editor/{registered_server_name}"
            )
            output["server_name"] = registered_server_name
        if cloudflare_token:
            output["cloudflare_token"] = cloudflare_token
        print_json(output)


@click.group()
def alph():
    """Launch Alph Editor from your local machine"""
    pass


@alph.command("launch")
@click.option("--port", type=int, default=8888, help="Port to run Jupyter Lab on")
@click.option("--no-browser", is_flag=True, help="Don't open browser automatically")
@click.option(
    "--directory", type=str, default=".", help="Directory to launch Jupyter Lab in"
)
@click.option(
    "--app-port",
    type=int,
    default=5000,
    help="Additional port to tunnel for your application (stored as port_forward_url)",
)
@click.option(
    "--api-key", envvar="AMDS_API_KEY", help="API key for dashboard integration"
)
@click.option("--allow-origin", type=str, default="*", help="CORS allow-origin setting")
@click.option(
    "--disable-sudo", is_flag=True, help="Disable sudo/root permissions in the notebook"
)
@click.option(
    "--output-format",
    type=click.Choice(["standard", "minimal", "json"]),
    default="standard",
    help="Output format for command results",
)
@click.option(
    "--thumbnail",
    type=str,
    default=None,
    help="Custom thumbnail URL for dashboard display",
)
@click.option(
    "--use-temp-tunnel",
    is_flag=True,
    help="Use temporary cloudflared tunnel instead of service",
)
@click.pass_obj
def launch_jupyter(
    client,
    port,
    no_browser,
    directory,
    app_port,
    api_key,
    allow_origin,
    disable_sudo,
    output_format,
    thumbnail,
    use_temp_tunnel,
):
    """Launch Alph Editor powered by your local Jupyter Lab server with Cloudflare proxy."""

    # Check prerequisites
    import shutil

    if not shutil.which("jupyter"):
        click.secho(
            "Error: Jupyter Lab is not installed. Please install it with:", fg="red"
        )
        click.echo("    pip install jupyterlab")
        sys.exit(1)

    if not shutil.which("cloudflared"):
        click.secho(
            "Error: cloudflared is not installed. Please install it from https://developers.cloudflare.com/cloudflare-one/connections/connect-networks/downloads/",
            fg="red",
        )
        sys.exit(1)

    # Check if port is in use and find an alternative if needed
    if is_port_in_use(port):
        click.secho(f"Warning: Port {port} is already in use.", fg="yellow")
        new_port = find_free_port(port + 1)
        if new_port:
            click.secho(f"Using alternative port {new_port} instead.", fg="yellow")
            port = new_port
        else:
            click.secho(
                "Error: Could not find an available port. Please specify a different port with --port.",
                fg="red",
            )
            sys.exit(1)

    # If app_port is specified, check if it's in use
    app_tunnel_url = None
    if app_port:
        if is_port_in_use(app_port):
            click.secho(f"Warning: App port {app_port} is already in use.", fg="yellow")
            new_app_port = find_free_port(app_port + 1)
            if new_app_port:
                click.secho(
                    f"Using alternative app port {new_app_port} instead.", fg="yellow"
                )
                app_port = new_app_port
            else:
                click.secho(
                    "Error: Could not find an available app port. Please specify a different port with --app-port.",
                    fg="red",
                )
                sys.exit(1)

    # Create a temporary directory for Jupyter config
    jupyter_dir = tempfile.mkdtemp(prefix="jupyter-amds-")

    try:
        # Create Jupyter config with CORS settings
        config_file = create_jupyter_config(jupyter_dir, allow_origin, disable_sudo)
        click.secho(
            f"Created Jupyter config with CORS allow-origin: {allow_origin}", fg="blue"
        )
        if disable_sudo:
            click.echo(
                "Root/sudo permissions have been disabled in the notebook server"
            )

        # Start Jupyter Lab process
        jupyter_process, jupyter_url, token = start_jupyter_server(
            directory, port, config_file
        )

        # Variables to store tunnel information
        tunnel_url = None
        app_tunnel_url = None
        cloudflare_token = None
        tunnel_process = None
        app_tunnel_process = None
        tunnel_name = None

        # Initialize client if needed
        if not client and api_key:
            from amds import Amds

            client = Amds(api_key=api_key)

        # Handle the tunneling - either tunnel or temporary
        if not use_temp_tunnel and (api_key or client):
            # Try to use the tunnel via API
            tunnel_info = setup_tunnel(client, port, app_port)

            if tunnel_info:
                tunnel_url = tunnel_info["tunnel_url"]
                app_tunnel_url = tunnel_info["app_tunnel_url"]
                cloudflare_token = tunnel_info["cloudflare_token"]
                tunnel_name = tunnel_info["tunnel_name"]
                tunnel_id = tunnel_info.get("tunnel_id")
                tunnel_name_for_dashboard = tunnel_name
            else:
                click.secho("Falling back to temporary tunnel...", fg="yellow")
                use_temp_tunnel = True
        else:
            # If requested or if we don't have API access, use temporary tunnel
            use_temp_tunnel = True

        # Set up temporary tunnel if needed
        if use_temp_tunnel:
            tunnel_info = setup_temporary_tunnel(port, app_port)

            if tunnel_info:
                tunnel_url = tunnel_info["tunnel_url"]
                tunnel_process = tunnel_info["tunnel_process"]
                app_tunnel_url = tunnel_info["app_tunnel_url"]
                app_tunnel_process = tunnel_info["app_tunnel_process"]
            else:
                jupyter_process.terminate()
                sys.exit(1)

        # Upload information to dashboard.amdatascience.com if we have an API key
        registered_server_name = None
        if api_key or client:
            if "tunnel_name_for_dashboard" in locals() and tunnel_name_for_dashboard:
                # Use the name from the tunnel API if available
                server_name = tunnel_name_for_dashboard
            else:
                # Generate a new server name
                server_name = f"local-{int(time.time())}"

            registered_server_name = register_with_dashboard(
                client,
                server_name,
                token,
                tunnel_url,
                app_tunnel_url,
                app_port,
                thumbnail,
            )
        else:
            click.echo(
                "Note: No API key provided. Running without dashboard integration."
            )

        # Open browser if requested
        if not no_browser:
            if registered_server_name:
                full_url = f"https://dashboard.amdatascience.com/alph-editor/{registered_server_name}"
            else:
                full_url = f"{jupyter_url}"

            click.echo(f"Opening browser at {full_url}")
            webbrowser.open(full_url)

        # Display information to user
        display_output(
            output_format,
            jupyter_url,
            tunnel_url,
            token,
            app_tunnel_url,
            registered_server_name,
            cloudflare_token if not use_temp_tunnel else None,
        )

        # Start health check thread
        health_thread = start_health_check_thread(
            jupyter_url, token, tunnel_url, app_tunnel_url
        )

        # Keep the process running until user interrupts
        try:
            jupyter_process.wait()
        except KeyboardInterrupt:
            click.secho("Shutting down...", fg="yellow")
        except Exception as e:
            click.secho(f"Error: {str(e)}", fg="red")
        finally:
            # Clean up processes
            try:
                jupyter_process.terminate()
                click.echo("Jupyter server stopped.")
            except Exception:
                pass

            try:
                if tunnel_process:
                    tunnel_process.terminate()
                    click.echo("Temporary tunnel closed.")
            except Exception:
                pass

            # Clean up app tunnel process if it exists
            if app_port and app_tunnel_process:
                try:
                    app_tunnel_process.terminate()
                    click.echo("Temporary app tunnel closed.")
                except Exception:
                    pass

            # If we used the tunnel service, try to clean it up
            if not use_temp_tunnel and cloudflare_token:
                cleanup_cloudflare_service()
                # Also clean up the tunnel on Cloudflare's side
                if tunnel_id and (api_key or client):
                    cleanup_tunnel_api(client, tunnel_id, api_key)

            # Delete the integrated server record if it was registered
            if registered_server_name and (api_key or client):
                cleanup_dashboard_integration(client, registered_server_name, api_key)

    except Exception as e:
        click.echo(f"Error: {str(e)}")
        sys.exit(1)
    finally:
        # Clean up temporary directory
        try:
            import shutil

            shutil.rmtree(jupyter_dir)
        except Exception:
            pass



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from amds.types import BaseModel
from amds.utils import FieldMetadata, QueryParamMetadata
from typing import Dict, List, Optional
from typing_extensions import Annotated, NotRequired, TypedDict


class DeleteTunnelRequestTypedDict(TypedDict):
    tunnel_id: NotRequired[str]


class DeleteTunnelRequest(BaseModel):
    tunnel_id: Annotated[
        Optional[str],
        FieldMetadata(query=QueryParamMetadata(style="form", explode=True)),
    ] = None


class DeleteTunnelResponseBodyTypedDict(TypedDict):
    r"""Delete Tunnel"""

    message: NotRequired[str]
    status: NotRequired[str]


class DeleteTunnelResponseBody(BaseModel):
    r"""Delete Tunnel"""

    message: Optional[str] = None

    status: Optional[str] = None


class DeleteTunnelResponseTypedDict(TypedDict):
    headers: Dict[str, List[str]]
    result: DeleteTunnelResponseBodyTypedDict


class DeleteTunnelResponse(BaseModel):
    headers: Dict[str, List[str]]

    result: DeleteTunnelResponseBody



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from amds.types import BaseModel
from amds.utils import FieldMetadata, PathParamMetadata, RequestMetadata
from typing import Dict, List, Optional
from typing_extensions import Annotated, NotRequired, TypedDict


class RunCodeRequestBodyTypedDict(TypedDict):
    code: NotRequired[str]
    kernel_name: NotRequired[str]


class RunCodeRequestBody(BaseModel):
    code: Optional[str] = None

    kernel_name: Optional[str] = None


class RunCodeRequestTypedDict(TypedDict):
    server_name: str
    request_body: NotRequired[RunCodeRequestBodyTypedDict]


class RunCodeRequest(BaseModel):
    server_name: Annotated[
        str, FieldMetadata(path=PathParamMetadata(style="simple", explode=False))
    ]

    request_body: Annotated[
        Optional[RunCodeRequestBody],
        FieldMetadata(request=RequestMetadata(media_type="application/json")),
    ] = None


class MetadataTypedDict(TypedDict):
    pass


class Metadata(BaseModel):
    pass


class OutputsTypedDict(TypedDict):
    name: NotRequired[str]
    output_type: NotRequired[str]
    text: NotRequired[str]


class Outputs(BaseModel):
    name: Optional[str] = None

    output_type: Optional[str] = None

    text: Optional[str] = None


class RunCodeContentTypedDict(TypedDict):
    cell_type: NotRequired[str]
    execution_count: NotRequired[float]
    metadata: NotRequired[MetadataTypedDict]
    outputs: NotRequired[List[OutputsTypedDict]]
    source: NotRequired[str]


class RunCodeContent(BaseModel):
    cell_type: Optional[str] = None

    execution_count: Optional[float] = None

    metadata: Optional[Metadata] = None

    outputs: Optional[List[Outputs]] = None

    source: Optional[str] = None


class RunCodeResponseBodyTypedDict(TypedDict):
    r"""Run Code"""

    content: NotRequired[RunCodeContentTypedDict]
    status: NotRequired[str]


class RunCodeResponseBody(BaseModel):
    r"""Run Code"""

    content: Optional[RunCodeContent] = None

    status: Optional[str] = None


class RunCodeResponseTypedDict(TypedDict):
    headers: Dict[str, List[str]]
    result: RunCodeResponseBodyTypedDict


class RunCodeResponse(BaseModel):
    headers: Dict[str, List[str]]

    result: RunCodeResponseBody



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from amds.types import BaseModel
from typing import Dict, List, Optional
from typing_extensions import NotRequired, TypedDict


class CreateTunnelRequestBodyTypedDict(TypedDict):
    app_port: NotRequired[float]
    jupyter_port: NotRequired[float]
    server_name: NotRequired[str]


class CreateTunnelRequestBody(BaseModel):
    app_port: Optional[float] = None

    jupyter_port: Optional[float] = None

    server_name: Optional[str] = None


class CreateTunnelDataTypedDict(TypedDict):
    app_url: NotRequired[str]
    created_at: NotRequired[str]
    hostname: NotRequired[str]
    id: NotRequired[str]
    jupyter_hostname: NotRequired[str]
    jupyter_url: NotRequired[str]
    name: NotRequired[str]
    token: NotRequired[str]


class CreateTunnelData(BaseModel):
    app_url: Optional[str] = None

    created_at: Optional[str] = None

    hostname: Optional[str] = None

    id: Optional[str] = None

    jupyter_hostname: Optional[str] = None

    jupyter_url: Optional[str] = None

    name: Optional[str] = None

    token: Optional[str] = None


class CreateTunnelResponseBodyTypedDict(TypedDict):
    r"""Create Tunnel"""

    data: NotRequired[CreateTunnelDataTypedDict]
    status: NotRequired[str]


class CreateTunnelResponseBody(BaseModel):
    r"""Create Tunnel"""

    data: Optional[CreateTunnelData] = None

    status: Optional[str] = None


class CreateTunnelResponseTypedDict(TypedDict):
    headers: Dict[str, List[str]]
    result: CreateTunnelResponseBodyTypedDict


class CreateTunnelResponse(BaseModel):
    headers: Dict[str, List[str]]

    result: CreateTunnelResponseBody



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from amds.types import BaseModel
from typing import Dict, List, Optional
from typing_extensions import NotRequired, TypedDict


class DataTypedDict(TypedDict):
    environment: NotRequired[str]
    port: NotRequired[float]
    port_forward_url: NotRequired[str]
    server_name: NotRequired[str]
    status: NotRequired[str]
    thumb: NotRequired[str]
    token: NotRequired[str]
    url: NotRequired[str]


class Data(BaseModel):
    environment: Optional[str] = None

    port: Optional[float] = None

    port_forward_url: Optional[str] = None

    server_name: Optional[str] = None

    status: Optional[str] = None

    thumb: Optional[str] = None

    token: Optional[str] = None

    url: Optional[str] = None


class GetIntegratedServersResponseBodyTypedDict(TypedDict):
    r"""Get Integrated Servers"""

    data: NotRequired[List[DataTypedDict]]
    status: NotRequired[str]


class GetIntegratedServersResponseBody(BaseModel):
    r"""Get Integrated Servers"""

    data: Optional[List[Data]] = None

    status: Optional[str] = None


class GetIntegratedServersResponseTypedDict(TypedDict):
    headers: Dict[str, List[str]]
    result: GetIntegratedServersResponseBodyTypedDict


class GetIntegratedServersResponse(BaseModel):
    headers: Dict[str, List[str]]

    result: GetIntegratedServersResponseBody



import click
import sys
import os
import json
from amds import Amds
from .commands import environments, servers, compute, login, alph
from .commands.login import get_config_path, is_token_valid


@click.group()
@click.option("--api-key", envvar="AMDS_API_KEY", help="API key for authentication")
@click.pass_context
def cli(ctx: click.Context, api_key: str):
    """American Data Science CLI - Command line interface for the American Data Science API"""
    
    # Skip API key check for login command
    if ctx.invoked_subcommand == 'login':
        ctx.obj = None
        return

    # If no API key provided via env or flag, try to get from config
    if not api_key:
        try:
            config_path = get_config_path()
            if os.path.exists(config_path):
                with open(config_path, "r", encoding="utf-8") as f:
                    config = json.load(f)
                    if is_token_valid(config):
                        api_key = config.get("auth_token")
        except Exception:
            pass

    if not api_key:
        click.echo(
            "Error: API key is required. Either:\n\n"
            "    1. Run 'amds login' to authenticate via browser\n"
            "    2. Set AMDS_API_KEY environment variable\n"
            "    3. Use --api-key option",
            err=True,
        )
        sys.exit(1)
    ctx.obj = Amds(api_key=api_key)


cli.add_command(environments.environments)
cli.add_command(servers.servers)
cli.add_command(compute.compute)
cli.add_command(login.login)
cli.add_command(alph.alph)  # Add the jupyter command

if __name__ == "__main__":
    # pylint: disable=no-value-for-parameter
    cli()



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from amds.types import BaseModel, Nullable, OptionalNullable, UNSET, UNSET_SENTINEL
from pydantic import model_serializer
from typing import Dict, List, Optional
from typing_extensions import NotRequired, TypedDict


class GetServersServersTypedDict(TypedDict):
    alph_editor_url: NotRequired[str]
    environment: NotRequired[str]
    name: NotRequired[str]
    port: NotRequired[float]
    port_forward_url: NotRequired[str]
    server_request: NotRequired[str]
    status: NotRequired[str]
    thumb: NotRequired[str]
    token: NotRequired[Nullable[str]]
    url: NotRequired[str]


class GetServersServers(BaseModel):
    alph_editor_url: Optional[str] = None

    environment: Optional[str] = None

    name: Optional[str] = None

    port: Optional[float] = None

    port_forward_url: Optional[str] = None

    server_request: Optional[str] = None

    status: Optional[str] = None

    thumb: Optional[str] = None

    token: OptionalNullable[str] = UNSET

    url: Optional[str] = None

    @model_serializer(mode="wrap")
    def serialize_model(self, handler):
        optional_fields = [
            "alph_editor_url",
            "environment",
            "name",
            "port",
            "port_forward_url",
            "server_request",
            "status",
            "thumb",
            "token",
            "url",
        ]
        nullable_fields = ["token"]
        null_default_fields = []

        serialized = handler(self)

        m = {}

        for n, f in type(self).model_fields.items():
            k = f.alias or n
            val = serialized.get(k)
            serialized.pop(k, None)

            optional_nullable = k in optional_fields and k in nullable_fields
            is_set = (
                self.__pydantic_fields_set__.intersection({n})
                or k in null_default_fields
            )  # pylint: disable=no-member

            if val is not None and val != UNSET_SENTINEL:
                m[k] = val
            elif val != UNSET_SENTINEL and (
                not k in optional_fields or (optional_nullable and is_set)
            ):
                m[k] = val

        return m


class GetServersResponseBodyTypedDict(TypedDict):
    r"""Get Servers"""

    servers: NotRequired[List[GetServersServersTypedDict]]
    status: NotRequired[str]


class GetServersResponseBody(BaseModel):
    r"""Get Servers"""

    servers: Optional[List[GetServersServers]] = None

    status: Optional[str] = None


class GetServersResponseTypedDict(TypedDict):
    headers: Dict[str, List[str]]
    result: GetServersResponseBodyTypedDict


class GetServersResponse(BaseModel):
    headers: Dict[str, List[str]]

    result: GetServersResponseBody



import click
from ..utils import print_json
import webbrowser


@click.group()
def servers():
    """Manage servers"""
    pass


@servers.command("list")
@click.pass_obj
def list_servers(client):
    """List all servers"""
    with client as c:
        res = c.servers.get()
        servers = res.result.servers
        # Print header
        click.echo(
            f"{'NAME':<20} {'ENVIRONMENT':<15} {'TYPE':<20} {'STATUS':<10} {'PORT'}"
        )
        click.echo("-" * 75)
        # Print each server's info
        for server in servers:
            click.echo(
                f"{server.name:<20} "
                f"{server.environment:<15} "
                f"{server.server_request:<20} "
                f"{server.status:<10} "
                f"{int(server.port)}"
            )


@servers.command("get")
@click.option("--server-name", required=True, help="Name of the server")
@click.pass_obj
def get_server(client, server_name):
    """Get server information"""
    with client as c:
        res = c.servers.get()
        servers = res.result.servers
        server = next((s for s in servers if s.name == server_name), None)
        if server:
            click.echo("\nServer Details:")
            click.echo("─" * 75)
            click.echo(f"Name:        {server.name}")
            click.echo(f"Environment: {server.environment}")
            click.echo(f"Type:        {server.server_request}")
            click.echo(f"Status:      {server.status}")
            click.echo(f"Port:        {int(server.port)}")
            if server.alph_editor_url:
                click.echo(
                    f"Alph Editor URL: {click.style(server.alph_editor_url, fg='bright_blue', underline=True)}"
                )
            if server.url:
                click.echo(
                    f"Jupyter Lab URL: {click.style(server.url, fg='bright_blue', underline=True)}"
                )
            if server.port_forward_url:
                click.echo(
                    f"Port Forward URL: {click.style(server.port_forward_url, fg='bright_blue', underline=True)}"
                )
            click.echo("─" * 75)
        else:
            click.echo(f"Server '{server_name}' not found", err=True)


@servers.command("create")
@click.option("--name", required=True, help="Name of the server")
@click.option("--environment", default="ai", help="Server environment")
@click.option("--port", type=int, default=5000, help="Server port")
@click.option("--server-type", default="amds-medium_cpu", help="Server type/request")
@click.pass_obj
def create_server(client, name, environment, port, server_type):
    """Create a server"""
    with client as c:
        res = c.servers.create(
            request={
                "environment": environment,
                "port": port,
                "server_name": name,
                "server_request": server_type,
            }
        )
        result = res.result

        click.echo(click.style("\nServer Created Successfully!", fg="green"))
        click.echo("─" * 75)
        click.echo(f"Name:        {result.server_name}")
        click.echo(f"Environment: {result.environment}")
        click.echo(f"Type:        {result.server_request}")
        click.echo(f"Port:        {int(result.port)}")
        click.echo(f"Status:      {result.status}")
        if result.hourly_rate:
            click.echo(f"Hourly Rate: ${result.hourly_rate/100:.2f}")

        if result.alph_editor_url:
            click.echo(
                f"\nAlph Editor URL: {click.style(result.alph_editor_url, fg='bright_blue', underline=True)}"
            )
        if result.url:
            click.echo(
                f"Jupyter Lab URL:\033]8;;{result.url}\033\\{result.url}\033]8;;\033\\"
            )
        if result.port_forward_url:
            click.echo(
                f"Forward URL: \033]8;;{result.port_forward_url}\033\\{result.port_forward_url}\033]8;;\033\\"
            )
        click.echo("─" * 75)


@servers.command("stop")
@click.option("--server-name", required=True, help="Name of the server")
@click.option("--force", is_flag=True, help="Skip confirmation prompt")
@click.option("--environment", required=False, default="ai", help="Server environment")
@click.option("--port", type=int, required=False, default=5000, help="Server port")
@click.option(
    "--server-type", required=False, default="amds-medium_cpu", help="Server type"
)
@click.pass_obj
def stop_server(client, server_name, force, environment, port, server_type):
    """Stop a server"""
    if not force and not click.confirm(f"Are you sure you want to stop server '{server_name}'?"):
        click.echo("Operation cancelled.")
        return
        
    with client as c:
        res = c.servers.stop(server_name=server_name)
        click.echo(click.style(f"\nServer '{server_name}' stopped successfully ✔", fg="green"))


@servers.command("start")
@click.option("--server-name", required=True, help="Name of the server")
@click.option("--environment", required=False, default="ai", help="Server environment")
@click.option("--port", type=int, required=False, default=5000, help="Server port")
@click.option(
    "--server-type", required=False, default="amds-medium_cpu", help="Server type"
)
@click.pass_obj
def start_server(client, server_name, environment, port, server_type):
    """Start a server"""
    with client as c:
        res = c.servers.start(
            server_name=server_name,
            environment=environment,
            port=port,
            server_request=server_type,
        )
        result = res.result

        click.echo(click.style("\nServer Started Successfully!", fg="green"))
        click.echo("─" * 75)
        click.echo(f"Name:        {result.server_name}")
        click.echo(f"Environment: {result.environment}")
        click.echo(f"Type:        {result.server_request}")
        click.echo(f"Port:        {int(result.port)}")
        click.echo(f"Status:      {result.status}")
        if result.hourly_rate:
            click.echo(f"Hourly Rate: ${result.hourly_rate/100:.2f}")

        if result.alph_editor_url:
            click.echo(f"\nAlph Editor URL: {result.alph_editor_url}")
        if result.url:
            click.echo(
                f"Jupyter Lab URL: \033]8;;{result.url}\033\\{result.url}\033]8;;\033\\"
            )
        if result.port_forward_url:
            click.echo(
                f"Port Forward URL: \033]8;;{result.port_forward_url}\033\\{result.port_forward_url}\033]8;;\033\\"
            )
        click.echo("─" * 75)


@servers.command("delete")
@click.option("--server-name", required=True, help="Name of the server")
@click.option("--force", is_flag=True, help="Skip confirmation prompt")
@click.pass_obj
def delete_server(client, server_name, force):
    """Delete a server"""
    if not force and not click.confirm(f"Are you sure you want to delete server '{server_name}'?"):
        click.echo("Operation cancelled.")
        return
        
    with client as c:
        res = c.servers.delete(server_name=server_name)
        click.echo(
            click.style(f"\nServer '{server_name}' deleted successfully ✔", fg="green")
        )


@servers.command("get-file", hidden=True)
@click.option("--server-name", required=True, help="Name of the server")
@click.option("--path", required=True, help="Remote file path")
@click.pass_obj
def get_file(client, server_name, path):
    """Get a file from a server"""
    with client as c:
        res = c.servers.get_file(server_name=server_name, path=path)
        print_json(res.model_dump())


@servers.command("upload-file", hidden=True)
@click.option("--server-name", required=True, help="Name of the server")
@click.option("--path", required=True, help="Remote file path")
@click.option("--content", help="File content")
@click.option("--format", "format_", default="text", help="File format")
@click.option("--type", "type_", default="file", help="Content type")
@click.pass_obj
def upload_file(client, server_name, path, content, format_, type_):
    """Upload a file to a server"""
    with client as c:
        res = c.servers.upload_file(
            server_name=server_name,
            path=path,
            content=content,
            format_=format_,
            type_=type_,
        )
        print_json(res.model_dump())


@servers.command("run-code", hidden=True)
@click.option("--server-name", required=True, help="Name of the server")
@click.option("--code", help="Python code to execute")
@click.option("--kernel-name", default="api-kernel", help="Kernel name")
@click.pass_obj
def run_code(client, server_name, code, kernel_name):
    """Run code on a server"""
    with client as c:
        res = c.servers.run_code(
            server_name=server_name, code=code, kernel_name=kernel_name
        )
        print_json(res.model_dump())


@servers.command("open")
@click.option("--server-name", required=True, help="Name of the server")
@click.pass_obj
def open_server_editor(client, server_name):
    """Open Alph Editor for a server in the default browser"""
    with client as c:
        res = c.servers.get()
        servers = res.result.servers
        server = next((s for s in servers if s.name == server_name), None)
        
        if server:
            # Check server status before proceeding
            if server.status.lower() not in ["running", "ready"]:
                click.echo(f"Server '{server_name}' is not running (status: {server.status})", err=True)
                click.echo("Please start the server before opening the editor.")
                return
                
            if server.alph_editor_url:
                click.echo(f"Opening Alph Editor for '{server_name}'...")
                webbrowser.open(server.alph_editor_url)
            else:
                click.echo(f"No Alph Editor URL available for server '{server_name}'", err=True)
        else:
            click.echo(f"Server '{server_name}' not found", err=True)



import click
import webbrowser
from http.server import HTTPServer, BaseHTTPRequestHandler
import urllib.parse
import threading
import json
import secrets
from typing import Optional
import os
import stat
from datetime import datetime, timedelta

# Store the authentication token
auth_token: Optional[str] = None

class AuthHandler(BaseHTTPRequestHandler):
    def do_GET(self):
        """Handle the callback from the authentication page"""
        self.send_response(404)
        self.end_headers()

    def do_POST(self):
        """Handle the POST callback from Auth0"""
        if self.path.startswith('/callback'):
            # Get the length of the POST data
            content_length = int(self.headers['Content-Length'])
            # Read the POST data
            post_data = self.rfile.read(content_length).decode('utf-8')
            # Parse the form data
            form_data = urllib.parse.parse_qs(post_data)
            
            # Get the token from the form data
            global auth_token
            auth_token = form_data.get('access_token', [None])[0]
            
            # Send response to browser
            self.send_response(200)
            self.send_header('Content-type', 'text/html')
            self.end_headers()
            
            response_content = """
            <html>
                <body>
                    <script>
                        window.close();
                        document.body.innerHTML = '<p>Authentication complete. You can close this window.</p>';
                    </script>
                </body>
            </html>
            """
            self.wfile.write(response_content.encode())
        else:
            self.send_response(404)
            self.end_headers()

def start_auth_server(port: int = 8081) -> HTTPServer:
    """Start the local authentication server"""
    server = HTTPServer(('localhost', port), AuthHandler)
    server_thread = threading.Thread(target=server.serve_forever)
    server_thread.daemon = True
    server_thread.start()
    return server

def get_config_path() -> str:
    """Get the path to the config directory and create it if it doesn't exist"""
    config_dir = os.path.expanduser("~/.config/amds")
    os.makedirs(config_dir, exist_ok=True)
    return os.path.join(config_dir, "config.json")

def is_token_valid(config: dict) -> bool:
    """Check if the stored token is still valid"""
    if 'expiration_time' not in config:
        return False
    
    try:
        expiration_time = datetime.fromisoformat(config['expiration_time'])
        return datetime.utcnow() < expiration_time
    except (ValueError, TypeError):
        return False

@click.command()
@click.option('--port', default=8081, help='Port for the authentication server')
def login(port: int):
    """Initiate the authentication process"""
    # Generate a state parameter for security
    state = secrets.token_urlsafe(16)
    
    # Auth0 configuration
    AUTH0_DOMAIN = "auth.amdatascience.com"
    AUTH0_CLIENT_ID = "FNaQyrcrgtlYSmBnASDKwMGHtezACTuC"
    # Ensure the redirect URI is properly formatted without any special characters
    REDIRECT_URI = f"http://localhost:{port}/callback".strip()
    
    # Construct the authentication URL for Auth0
    auth_url = (
        f"https://{AUTH0_DOMAIN}/authorize"
        f"?response_type=token"
        f"&response_mode=form_post"
        f"&client_id={AUTH0_CLIENT_ID}"
        f"&redirect_uri={REDIRECT_URI}"  # Remove urllib.parse.quote since Auth0 will handle the encoding
        f"&scope=openid profile email"
        f"&state={state}"
    )
    
    # Start the local server
    server = start_auth_server(port)
    
    # Open the browser
    click.echo("Opening browser for authentication...")
    webbrowser.open(auth_url)
    
    # Wait for authentication to complete
    while auth_token is None:
        click.echo("Waiting for authentication...", nl=False)
        click.echo("\r", nl=False)
    
    # Shutdown the server
    server.shutdown()
    server.server_close()
    
    # Store the token securely
    if auth_token:
        click.echo("Authentication successful!")
        config_path = get_config_path()
        
        # Set expiration time to 12 hours from now
        expiration_time = datetime.utcnow() + timedelta(hours=12)
        
        config = {
            'auth_token': auth_token,
            'expiration_time': expiration_time.isoformat()
        }
        
        # Write config with restricted permissions (600)
        with open(config_path, 'w') as f:
            json.dump(config, f)
        
        # Set file permissions to user read/write only
        os.chmod(config_path, stat.S_IRUSR | stat.S_IWUSR)
        
        click.echo(f"Token stored in {config_path}")
    else:
        click.echo("Authentication failed!")



import click
from ..utils import print_json

@click.group()
def environments():
    """Manage environments"""
    pass

@environments.command('list')
@click.pass_obj
def list_environments(client):
    """List all environments"""
    with client as c:
        res = c.environments.get()

        result = res.result
        
        if not result.environments:
            click.echo("No environments found.")
            return

        click.echo("\nAvailable Environments:")
        click.echo("─" * 75)
        
        for env in result.environments:
            click.echo(f"Name:        {click.style(env.name, fg='bright_green')}")
            click.echo(f"ID:          {env.environment_id}")
            if hasattr(env, "description") and env.description:
                click.echo(f"Description: {env.description}")
            click.echo("─" * 75) 

@environments.command('get')
@click.option("--environment-id", required=True, help="ID of the environment")
@click.pass_obj
def get_environment(client, environment_id):
    """Get details of a specific environment by name"""
    with client as c:
        res = c.environments.get()
        result = res.result
        
        if not result.environments:
            click.echo("No environments found.")
            return
            
        environment = next((env for env in result.environments if env.environment_id == environment_id), None)
        
        if environment:
            click.echo("\nEnvironment Details:")
            click.echo("─" * 75)
            click.echo(f"Name:        {click.style(environment.name, fg='bright_green')}")
            click.echo(f"ID:          {environment.environment_id}")
            if hasattr(environment, "description") and environment.description:
                click.echo(f"Description: {environment.description}")
            click.echo("─" * 75)
        else:
            click.echo(f"Environment '{name}' not found", err=True) 


import click
from ..utils import print_json

@click.group()
def compute():
    """Manage compute resources"""
    pass

@compute.command('list')
@click.pass_obj
def list_compute(client):
    """List compute information"""
    with client as c:
        res = c.compute.get()

        result = res.result
        
        if not result.compute:
            click.echo("No compute resources found.")
            return
            
        click.echo("\nAvailable Compute Resources:")
        click.echo("─" * 75)
        
        for comp in result.compute:
            click.echo(f"Name:         {click.style(comp.name, fg='bright_green')}")
            click.echo(f"Host:         {comp.cloud_host}")
            if hasattr(comp, "memory"):
                click.echo(f"Memory:       {comp.memory}")
            if hasattr(comp, "cpus"):
                click.echo(f"CPUs:         {comp.cpus}")
            if hasattr(comp, "num_gpus"):
                click.echo(f"GPUs:         {comp.num_gpus}")
            if hasattr(comp, "vram_per_gpu") and comp.vram_per_gpu != "0G":
                click.echo(f"VRAM/GPU:     {comp.vram_per_gpu}")
            if hasattr(comp, "total_vram") and comp.total_vram != "0G":
                click.echo(f"Total VRAM:   {comp.total_vram}")
            if hasattr(comp, "hourly_price"):
                click.echo(f"Hourly Price: ${comp.hourly_price}")
                
            # Display availability information
            if hasattr(comp, "availability") and comp.availability:
                click.echo("Availability: ")
                for region in comp.availability:
                    status = click.style("Available", fg="green") if region.available else click.style("Unavailable", fg="red")
                    click.echo(f"  - {region.region}: {status}")
                    
            click.echo("─" * 75) 

@compute.command('get')
@click.option("--compute-id", required=True, help="ID of the compute resource")
@click.pass_obj
def get_compute(client, compute_id):
    """Get details of a specific compute resource by name"""
    with client as c:
        res = c.compute.get()
        result = res.result
        
        if not result.compute:
            click.echo("No compute resources found.")
            return
            
        compute = next((comp for comp in result.compute if comp.compute_id== compute_id), None)
        
        if compute:
            click.echo("\nCompute Resource Details:")
            click.echo("─" * 75)
            click.echo(f"Name:         {click.style(compute.name, fg='bright_green')}")
            click.echo(f"Host:         {compute.cloud_host}")
            if hasattr(compute, "memory"):
                click.echo(f"Memory:       {compute.memory}")
            if hasattr(compute, "cpus"):
                click.echo(f"CPUs:         {compute.cpus}")
            if hasattr(compute, "num_gpus"):
                click.echo(f"GPUs:         {compute.num_gpus}")
            if hasattr(compute, "vram_per_gpu") and compute.vram_per_gpu != "0G":
                click.echo(f"VRAM/GPU:     {compute.vram_per_gpu}")
            if hasattr(compute, "total_vram") and compute.total_vram != "0G":
                click.echo(f"Total VRAM:   {compute.total_vram}")
            if hasattr(compute, "hourly_price"):
                click.echo(f"Hourly Price: ${compute.hourly_price}")
                
            # Display availability information
            if hasattr(compute, "availability") and compute.availability:
                click.echo("Availability: ")
                for region in compute.availability:
                    status = click.style("Available", fg="green") if region.available else click.style("Unavailable", fg="red")
                    click.echo(f"  - {region.region}: {status}")
                    
            click.echo("─" * 75)
        else:
            click.echo(f"Compute resource '{name}' not found", err=True) 